@@ -25,11 +25,11 @@ in [SageMath](https://www.sagemath.org/) and
 [Maxima](http://maxima.sourceforge.net/).
 """
 
+from sympy.core.add import _unevaluated_Add
 from sympy.core.expr import Expr
 from sympy.core.basic import Basic
 from sympy.core.evalf import EvalfMixin
 from sympy.core.sympify import _sympify
-from sympy import functions
 import functools
 from sympy import *
 
@@ -146,6 +146,10 @@ class Equation(Basic, EvalfMixin):
 
     Examples
     ========
+    NOTE: All the examples below are in vanilla python. You can get human
+    readable eqautions "lhs = rhs" in vanilla python by adjusting the settings
+    in `algwsym_config` (see it's documentation). Output is human readable by
+    default in IPython and Jupyter environments.
     >>> from algebra_with_sympy import *
     >>> a, b, c, x = var('a b c x')
     >>> Equation(a,b/c)
@@ -251,6 +255,21 @@ class Equation(Basic, EvalfMixin):
     >>> eq2.subs({R:0.08206*L*atm/mol/K,T:273*K,n:1.00*mol,V:24.0*L}).evalf(4)
     Equation(p, 0.9334*atm)
 
+    Substituting an equation into another equation:
+    >>> P, P1, P2, A1, A2, E1, E2 = symbols("P, P1, P2, A1, A2, E1, E2")
+    >>> eq1 = Eqn(P, P1 + P2)
+    >>> eq2 = Eqn(P1 / (A1 * E1), P2 / (A2 * E2))
+    >>> P1_val = (eq1 - P2).swap
+    >>> P1_val
+    Equation(P1, P - P2)
+    >>> eq2 = eq2.subs(P1_val)
+    >>> eq2
+    Equation((P - P2)/(A1*E1), P2/(A2*E2))
+    >>> P2_val = solve(eq2.subs(P1_val), P2)[0]
+    P2 = A2*E2*P/(A1*E1 + A2*E2)
+    >>> P2_val
+    Equation(P2, A2*E2*P/(A1*E1 + A2*E2))
+
     Combining equations (Math with equations: lhs with lhs and rhs with rhs)
     >>> q = Eqn(a*c, b/c**2)
     >>> q
@@ -327,15 +346,19 @@ class Equation(Basic, EvalfMixin):
     >>> q.dorhs.integrate(b).dolhs.integrate(a)
     Equation(a**2*c/2, b**2/(2*c))
 
-    SymPy's solvers do not understand these equations. They expect an
-    expression that the solver assumes = 0. Thus to use the solver the
-    equation must be rearranged so that all non-zero symbols are on one side.
-    Then just the non-zero symbolic side is passed to ``solve()``.
-    >>> t2 = t-t.rhs
-    >>> t2
-    Equation(a - b/c, 0)
-    >>> solve(t2.lhs,c)
-    [b/a]
+    Automatic solutions using sympy solvers. THIS IS EXPERIMENTAL. Please
+    report issues at https://github.com/gutow/Algebra_with_Sympy/issues.
+    >>> tosolv = Eqn(a - b, c/a)
+    >>> solve(tosolv,a)
+    a = b/2 - sqrt(b**2 + 4*c)/2
+    a = b/2 + sqrt(b**2 + 4*c)/2
+    [Equation(a, b/2 - sqrt(b**2 + 4*c)/2), Equation(a, b/2 + sqrt(b**2 + 4*c)/2)]
+    >>> solve(tosolv, b)
+    b = (a**2 - c)/a
+    [Equation(b, (a**2 - c)/a)]
+    >>> solve(tosolv, c)
+    c = a**2 - a*b
+    [Equation(c, a**2 - a*b)]
 
     """
 
@@ -347,17 +370,22 @@ class Equation(Basic, EvalfMixin):
         return super().__new__(cls, lhs, rhs)
 
     def _get_eqn_name(self):
-        from IPython import get_ipython
-        global_dict = getattr(get_ipython(), 'user_ns', None)
+        """
+        Tries to find the python string name that refers to the equation. In
+        IPython environments (IPython, Jupyter, etc...) looks in the user_ns.
+        If not in an IPython environment looks in __main__.
+        :return: string value if found or empty string.
+        """
         human_text = algwsym_config.output.human_text
         algwsym_config.output.human_text=False
-        if global_dict:
-            for var_name in global_dict:
-                if isinstance(global_dict[var_name], Equation):
-                    if (global_dict[var_name]).__repr__()==self.__repr__() and not \
-                            var_name.startswith('_'):
-                        algwsym_config.output.human_text=human_text
-                        return var_name
+        import __main__ as shell
+        for k in dir(shell):
+            item = getattr(shell,k)
+            if isinstance(item,Equation):
+                if item.__repr__()==self.__repr__() and not \
+                        k.startswith('_'):
+                    algwsym_config.output.human_text=human_text
+                    return k
         algwsym_config.output.human_text = human_text
         return ''
 
@@ -379,7 +407,6 @@ class Equation(Basic, EvalfMixin):
         """
         Converts the equation to an Equality.
         """
-        from sympy.core.relational import Equality
         return Equality(self.lhs, self.rhs)
 
     def check(self, **kwargs):
@@ -393,7 +420,6 @@ class Equation(Basic, EvalfMixin):
         -------
         True, False or an unevaluated `Equality` if truth cannot be determined.
         """
-        from sympy.core.relational import Equality
         return Equality(self.lhs, self.rhs, **kwargs).simplify()
 
     @property
@@ -501,6 +527,135 @@ class Equation(Basic, EvalfMixin):
     @property
     def dorhs(self):
         return self._sides(self, side='rhs')
+    
+    def _eval_rewrite(self, rule, args, **kwargs):
+        """Return Equation(L, R) as Equation(L - R, 0) or as L - R.
+
+        Parameters
+        ==========
+
+        evaluate : bool, optional
+            Control the evaluation of the result. If `evaluate=None` then
+            terms in L and R will not cancel but they will be listed in
+            canonical order; otherwise non-canonical args will be returned.
+            Default to True.
+        
+        eqn : bool, optional
+            Control the returned type. If `eqn=True`, then Equation(L - R, 0)
+            is returned. Otherwise, the L - R symbolic expression is returned.
+            Default to True.
+
+        Examples
+        ========
+        >>> from sympy import Add
+        >>> from sympy.abc import b, x
+        >>> from algebra_with_sympy import Equation
+        >>> eq = Equation(x + b, x - b)
+        >>> eq.rewrite(Add)
+        Equation(2*b, 0)
+        >>> eq.rewrite(Add, evaluate=None).lhs.args
+        (b, b, x, -x)
+        >>> eq.rewrite(Add, evaluate=False).lhs.args
+        (b, x, b, -x)
+        >>> eq.rewrite(Add, eqn=False)
+        2*b
+        >>> eq.rewrite(Add, eqn=False, evaluate=False).args
+        (b, x, b, -x)
+        """
+        if rule == Add:
+            # NOTE: the code about `evaluate` is very similar to
+            # sympy.core.relational.Equality._eval_rewrite_as_Add
+            eqn = kwargs.pop("eqn", True)
+            evaluate = kwargs.get('evaluate', True)
+            L, R = args
+            if evaluate:
+                # allow cancellation of args
+                expr = L - R
+            else:
+                args = Add.make_args(L) + Add.make_args(-R)
+                if evaluate is None:
+                    # no cancellation, but canonical
+                    expr = _unevaluated_Add(*args)
+                else:
+                    # no cancellation, not canonical
+                    expr = Add._from_args(args)
+            if eqn:
+                return self.func(expr, 0)
+            return expr
+
+    def subs(self, *args, **kwargs):
+        """Substitutes old for new in an equation after sympifying args.
+    
+        `args` is either:
+
+        * one or more arguments of type `Equation(old, new)`.
+        * two arguments, e.g. foo.subs(old, new)
+        * one iterable argument, e.g. foo.subs(iterable). The iterable may be:
+
+            - an iterable container with (old, new) pairs. In this case the
+              replacements are processed in the order given with successive
+              patterns possibly affecting replacements already made.
+            - a dict or set whose key/value items correspond to old/new pairs.
+              In this case the old/new pairs will be sorted by op count and in
+              case of a tie, by number of args and the default_sort_key. The
+              resulting sorted list is then processed as an iterable container
+              (see previous).
+        
+        If the keyword ``simultaneous`` is True, the subexpressions will not be
+        evaluated until all the substitutions have been made.
+
+        Please, read ``help(Expr.subs)`` for more examples.
+
+        Examples
+        ========
+
+        >>> from sympy.abc import a, b, c, x
+        >>> from algebra_with_sympy import Equation
+        >>> eq = Equation(x + a, b * c)
+
+        Substitute a single value:
+
+        >>> eq.subs(b, 4)
+        Equation(a + x, 4*c)
+
+        Substitute a multiple values:
+
+        >>> eq.subs([(a, 2), (b, 4)])
+        Equation(x + 2, 4*c)
+        >>> eq.subs({a: 2, b: 4})
+        Equation(x + 2, 4*c)
+
+        Substitute an equation into another equation:
+
+        >>> eq2 = Equation(x + a, 4)
+        >>> eq.subs(eq2)
+        Equation(4, b*c)
+
+        Substitute multiple equations into another equation:
+
+        >>> eq1 = Equation(x + a + b + c, x * a * b * c)
+        >>> eq2 = Equation(x + a, 4)
+        >>> eq3 = Equation(b, 5)
+        >>> eq1.subs(eq2, eq3)
+        Equation(c + 9, 5*a*c*x)
+
+        """
+        new_args = args
+        if all(isinstance(a, self.func) for a in args):
+            new_args = [{a.args[0]: a.args[1] for a in args}]
+        elif (len(args) == 1) and all(isinstance(a, self.func) for a in
+                                      args[0]):
+            raise TypeError("You passed into `subs` a list of elements of "
+                "type `Equation`, but this is not supported. Please, consider "
+                "unpacking the list with `.subs(*eq_list)` or select your "
+                "equations from the list and use `.subs(eq_list[0], eq_list["
+                "2], ...)`.")
+        elif any(isinstance(a, self.func) for a in args):
+            raise ValueError("`args` contains one or more Equation and some "
+                "other data type. This mode of operation is not supported. "
+                "Please, read `subs` documentation to understand how to "
+                "use it.")
+        return super().subs(*new_args, **kwargs)
 
     #####
     # Overrides of binary math operations
@@ -627,14 +782,14 @@ class Equation(Basic, EvalfMixin):
             return self.__str__()
         return repstr
 
-    def _latex(self, obj, **kwargs):
+    def _latex(self, printer):
         tempstr = ''
         if algwsym_config.output.show_code and not \
             algwsym_config.output.human_text:
-            tempstr +='\\text{code version: '+ self.__repr__()+'} \\newline '
-        tempstr += latex(self.lhs, **kwargs)
+            print('code version: '+ self.__repr__())
+        tempstr += printer._print(self.lhs)
         tempstr += '='
-        tempstr += latex(self.rhs, **kwargs)
+        tempstr += printer._print(self.rhs)
         namestr = self._get_eqn_name()
         if namestr !='' and algwsym_config.output.label:
             tempstr += '\\,\\,\\,\\,\\,\\,\\,\\,\\,\\,'
@@ -685,12 +840,24 @@ def solve(f, *symbols, **flags):
     flags['dict'] = True
     result = solve(newf, *symbols, **flags)
     if contains_eqn:
-        for k in result:
-            for key in k.keys():
-                val = k[key]
-                tempeqn = Eqn(key, val)
-                display(tempeqn)
-                solns.append(tempeqn)
+        if len(result[0]) == 1:
+            for k in result:
+                for key in k.keys():
+                    val = k[key]
+                    tempeqn = Eqn(key, val)
+                    display(tempeqn)
+                    solns.append(tempeqn)
+        else:
+            for k in result:
+                solnset = []
+                for key in k.keys():
+                    val = k[key]
+                    tempeqn = Eqn(key, val)
+                    display(tempeqn)
+                    solnset.append(tempeqn)
+                print('-----')
+                solns.append(solnset)
+
     else:
         solns = result
     return solns
@@ -718,14 +885,17 @@ del symsqrt
 def root(arg, n, k = 0, evaluate = None):
     """
     Override of sympy convenience function `root`. Simply divides equations
-    into two sides if `arg` is an instance of `Equation`. This avoids an
-    issue with the way sympy is delaying specialized applications of _Pow_ on
-    objects that are not basic sympy expressions.
+    into two sides if `arg`  or `n` is an instance of `Equation`. This
+    avoids an issue with the way sympy is delaying specialized applications
+    of _Pow_ on objects that are not basic sympy expressions.
     """
     from sympy.functions.elementary.miscellaneous import root as symroot
     if isinstance(arg, Equation):
-        return Equation(symroot(arg.lhs, n, k, evaluate), symroot(arg.rhs,
-                                                            n, k, evaluate))
+        return Equation(symroot(arg.lhs, n, k, evaluate),
+                        symroot(arg.rhs, n, k, evaluate))
+    if isinstance(n, Equation):
+        return Equation(symroot(arg, n.lhs, k, evaluate),
+                        symroot(arg, n.rhs, k, evaluate))
     else:
         return symroot(arg, n, k, evaluate)
 
@@ -734,6 +904,23 @@ from sympy.functions.elementary.miscellaneous import root as symroot
 root.__doc__+=symroot.__doc__
 del symroot
 
+def Heaviside(arg, **kwargs):
+    """
+    Overide of the Heaviside function as implemented in Sympy. Get a recursion
+    error if use the normal class extension of a function to do this.
+
+    """
+    from sympy.functions.special.delta_functions import Heaviside as symHeav
+    if isinstance(arg, Equation):
+        return Equation(symHeav((arg.lhs), **kwargs),symHeav((arg.rhs),
+                                                             **kwargs))
+    else:
+        return symHeav(arg, **kwargs)
+# Pick up the docstring for Heaviside from Sympy.
+from sympy.functions.special.delta_functions import Heaviside as symHeav
+Heaviside.__doc__ += symHeav.__doc__
+del symHeav
+
 def collect(expr, syms, func=None, evaluate=None, exact=False,
             distribute_order_term=True):
     """
@@ -748,7 +935,7 @@ def collect(expr, syms, func=None, evaluate=None, exact=False,
         return collect(expr, syms, func, evaluate, exact,
                        distribute_order_term)
 
-class Eq(Eq):
+class Equality(Equality):
     """
     Extension of Equality class to include the ability to convert it to an
     Equation.
@@ -766,6 +953,7 @@ class Eq(Eq):
         """
         return self.to_Equation()
 
+Eq = Equality
 #####
 # Extension of the Function class. For incorporation into SymPy this should
 # become part of the class
@@ -802,26 +990,50 @@ class EqnFunction(Function):
                 return Equation(lhs,rhs)
         return super().__new__(cls, *args, **kwargs)
 
+def str_to_extend_sympy_func(func:str):
+    """
+    Generates the string command to execute for a sympy function to
+    gain the properties of the extended EqnFunction class.
+    """
+    execstr = 'class ' + str(func) + '(' + str(
+        func) + ',EqnFunction):\n    ' \
+                'pass\n'
+    return execstr
+
+# TODO: Below will not be needed when incorporated into SymPy.
+# This is hacky, but I have not been able to come up with another way
+# of extending the functions programmatically, if this is separate package
+# from sympy that extends it after loading sympy.
+#  Functions listed in `skip` are not applicable to equations or cannot be
+#  extended because of `mro` error or `metaclass conflict`. This reflects
+#  that some of these are not members of the Sympy Function class.
+
+# Overridden elsewhere
+_extended_ = ('sqrt', 'root', 'Heaviside')
+
+# Either not applicable to equations or have not yet figured out a way
+# to systematically apply to an equation.
+# TODO examine these more carefully (top priority: real_root, cbrt, Ynm_c).
+_not_applicable_to_equations_ = ('Min', 'Max', 'Id', 'real_root', 'cbrt',
+        'unbranched_argument', 'polarify', 'unpolarify',
+        'piecewise_fold', 'E1', 'Eijk', 'bspline_basis',
+        'bspline_basis_set', 'interpolating_spline', 'jn_zeros',
+        'jacobi_normalized', 'Ynm_c', 'piecewise_exclusive', 'Piecewise',
+        'motzkin', 'hyper','meijerg', 'chebyshevu_root', 'chebyshevt_root',
+        'betainc_regularized')
+_skip_ = _extended_ + _not_applicable_to_equations_
 
 for func in functions.__all__:
-    # TODO: This will not be needed when incorporated into SymPy listed in
-    #  `skip` cannot be extended because of `mro` error or `metaclass
-    #  conflict`. Seems to reflect expectation that a helper function will be
-    #  defined within the object (e.g. `_eval_power()` for all the flavors of
-    #  `root`).
-    skip = ('sqrt', 'root', 'Min', 'Max', 'Id', 'real_root', 'cbrt',
-            'unbranched_argument', 'polarify', 'unpolarify',
-            'piecewise_fold', 'E1', 'Eijk', 'bspline_basis',
-            'bspline_basis_set', 'interpolating_spline', 'jn_zeros',
-            'jacobi_normalized', 'Ynm_c', 'piecewise_exclusive')
-    if func not in skip:
+
+    if func not in _skip_:
         try:
-            execstr = 'class ' + str(func) + '(' + str(
-                func) + ',EqnFunction):\n    pass\n'
-            exec(execstr, globals(), locals())
+            exec(str_to_extend_sympy_func(func), globals(), locals())
         except TypeError:
             from warnings import warn
-            warn('SymPy function/operation '+str(func)+ ' may not work ' \
-                 'properly with Equations. If you use it with Equations, ' \
-                 'validate its behavior. We are working to address this ' \
-                 'issue.')
\ No newline at end of file
+            warn('SymPy function/operation ' + str(func) + ' may not work ' \
+                'properly with Equations. If you use it with Equations, ' \
+                'validate its behavior. We are working to address this ' \
+                'issue.')
+
+# Redirect python abs() to Abs()
+abs = Abs
\ No newline at end of file
